@@ -6,6 +6,7 @@ from __future__ import absolute_import, division, print_function
 import io
 from glob import has_magic
 import logging
+import os
 import warnings
 
 from azure.core.exceptions import ResourceNotFoundError, ResourceExistsError
@@ -14,15 +15,16 @@ from azure.datalake.store import AzureDLFileSystem, lib
 from azure.datalake.store.core import AzureDLFile, AzureDLPath
 from azure.storage.blob.aio import BlobServiceClient as AIOBlobServiceClient
 from azure.storage.blob import BlobServiceClient
-from azure.storage.blob.aio._models import BlobPrefix
-from azure.storage.blob._models import BlobBlock, BlobProperties
+from azure.storage.blob.aio._list_blobs_helper import BlobPrefix
+from azure.storage.blob._models import BlobBlock, BlobProperties, BlobType
 from fsspec import AbstractFileSystem
 from fsspec.asyn import (
-    sync,
+    maybe_sync,
     AsyncFileSystem,
     get_loop,
+    sync_wrapper,
 )
-from fsspec.utils import infer_storage_options, tokenize, other_paths
+from fsspec.utils import infer_storage_options, tokenize
 
 
 logger = logging.getLogger(__name__)
@@ -429,25 +431,29 @@ class AzureBlobFileSystem(AsyncFileSystem):
         ------
         ValueError if none of the connection details are available
         """
-        self.account_url: str = f"https://{self.account_name}.blob.core.windows.net"
-        if self.credential is not None:
-            self.service_client = AIOBlobServiceClient(
-                account_url=self.account_url, credential=self.credential
-            )
-        elif self.connection_string is not None:
-            self.service_client = AIOBlobServiceClient.from_connection_string(
-                conn_str=self.connection_string
-            )
-        elif self.account_key is not None:
-            self.service_client = AIOBlobServiceClient(
-                account_url=self.account_url, credential=self.account_key
-            )
-        elif self.sas_token is not None:
-            self.service_client = AIOBlobServiceClient(
-                account_url=self.account_url + self.sas_token, credential=None
-            )
-        else:
-            raise ValueError("unable to connect with provided params!!")
+        try:
+            self.account_url: str = f"https://{self.account_name}.blob.core.windows.net"
+            if self.credential is not None:
+                self.service_client = AIOBlobServiceClient(
+                    account_url=self.account_url, credential=self.credential
+                )
+            elif self.connection_string is not None:
+                self.service_client = AIOBlobServiceClient.from_connection_string(
+                    conn_str=self.connection_string
+                )
+            elif self.account_key is not None:
+                self.service_client = AIOBlobServiceClient(
+                    account_url=self.account_url, credential=self.account_key
+                )
+            elif self.sas_token is not None:
+                self.service_client = AIOBlobServiceClient(
+                    account_url=self.account_url + self.sas_token, credential=None
+                )
+            else:
+                self.service_client = AIOBlobServiceClient(account_url=self.account_url)
+
+        except Exception as e:
+            raise ValueError(f"unable to connect to account for {e}")
 
     def split_path(self, path, delimiter="/", return_container: bool = False, **kwargs):
         """
@@ -481,13 +487,9 @@ class AzureBlobFileSystem(AsyncFileSystem):
             return path.split(delimiter, 1)
 
     def info(self, path, refresh=False, **kwargs):
-        # print(path)
-        # print(self._ls_from_cache(""))
-        # print(refresh)
         fetch_from_azure = (path and self._ls_from_cache(path) is None) or refresh
-        # print(fetch_from_azure)
         if fetch_from_azure:
-            return sync(self.loop, self._info, path)
+            return maybe_sync(self._info, self, path)
         return super().info(path)
 
     async def _info(self, path, **kwargs):
@@ -521,7 +523,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
             raise FileNotFoundError
 
     def glob(self, path, **kwargs):
-        return sync(self.loop, self._glob, path)
+        return maybe_sync(self._glob, self, path)
 
     async def _glob(self, path, **kwargs):
         """
@@ -608,9 +610,9 @@ class AzureBlobFileSystem(AsyncFileSystem):
         **kwargs,
     ):
 
-        files = sync(
-            self.loop,
+        files = maybe_sync(
             self._ls,
+            self,
             path=path,
             invalidate_cache=invalidate_cache,
             delimiter=delimiter,
@@ -769,7 +771,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         return output
 
     def find(self, path, maxdepth=None, withdirs=False, **kwargs):
-        return sync(self.loop, self._find, path, maxdepth, withdirs)
+        return maybe_sync(self._find, self, path, maxdepth, withdirs)
 
     async def _find(self, path, maxdepth=None, withdirs=False, **kwargs):
         """List all files below path.
@@ -876,10 +878,10 @@ class AzureBlobFileSystem(AsyncFileSystem):
             ):
                 yield path, dirs, files
 
-    def mkdir(self, path, delimiter="/", exists_ok=False, **kwargs):
-        sync(self.loop, self._mkdir, path, delimiter, exists_ok)
+    def mkdir(self, path, delimiter="/", exist_ok=False, **kwargs):
+        maybe_sync(self._mkdir, self, path, delimiter, exist_ok)
 
-    async def _mkdir(self, path, delimiter="/", exists_ok=False, **kwargs):
+    async def _mkdir(self, path, delimiter="/", exist_ok=False, **kwargs):
         """
         Create directory entry at path
 
@@ -891,7 +893,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         delimiter: str
             Delimiter to use when splitting the path
 
-        exists_ok: bool
+        exist_ok: bool
             If True, raise an exception if the directory already exists. Defaults to False
         """
         container_name, path = self.split_path(path, delimiter=delimiter)
@@ -903,7 +905,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         container_name_as_dir = f"{container_name}/"
         if _containers is None:
             _containers = []
-        if not exists_ok:
+        if not exist_ok:
             if (container_name_as_dir not in _containers) and (not path):
                 # create new container
                 await self.service_client.create_container(name=container_name)
@@ -928,10 +930,11 @@ class AzureBlobFileSystem(AsyncFileSystem):
                     await container_client.upload_blob(name=path, data="")
             except ResourceExistsError:
                 pass
-        _ = await self._ls("", invalidate_cache=True)
+        # _ = await self._ls("", invalidate_cache=True)
+        self.invalidate_cache(self._parent(path))
 
     def rm(self, path, recursive=False, maxdepth=None, **kwargs):
-        sync(self.loop, self._rm, path, recursive, **kwargs)
+        maybe_sync(self._rm, self, path, recursive, **kwargs)
 
     async def _rm(self, path, recursive=False, maxdepth=None, **kwargs):
         """Delete files.
@@ -950,6 +953,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         path = await self._expand_path(path, recursive=recursive, maxdepth=maxdepth)
         for p in reversed(path):
             await self.rm_file(p)
+        self.invalidate_cache()
 
     async def rm_file(self, path, delimiter="/", **kwargs):
         """
@@ -985,13 +989,13 @@ class AzureBlobFileSystem(AsyncFileSystem):
                     await container_client.delete_container()
             else:
                 raise RuntimeError(f"Unable to delete {path}!")
-            _ = await self._ls("", invalidate_cache=True)
+            self.invalidate_cache(self._parent(path))
 
         except FileNotFoundError:
             pass
 
     def rmdir(self, path: str, delimiter="/", **kwargs):
-        sync(self.loop, self._rmdir, path, delimiter=delimiter, **kwargs)
+        maybe_sync(self._rmdir, self, path, delimiter=delimiter, **kwargs)
 
     async def _rmdir(self, path: str, delimiter="/", **kwargs):
         """
@@ -1013,10 +1017,10 @@ class AzureBlobFileSystem(AsyncFileSystem):
         if (container_name + delimiter in _containers) and (not path):
             # delete container
             await self.service_client.delete_container(container_name)
-            _ = await self._ls("", invalidate_cache=True)
+            self.invalidate_cache(self._parent(path))
 
     def size(self, path):
-        return sync(self.loop, self._size, path)
+        return maybe_sync(self._size, self, path)
 
     async def _size(self, path):
         """Size in bytes of file"""
@@ -1025,7 +1029,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         return size
 
     def isfile(self, path):
-        return sync(self.loop, self._isfile, path)
+        return maybe_sync(self._isfile, self, path)
 
     async def _isfile(self, path):
         """Is this entry file-like?"""
@@ -1045,7 +1049,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
             return False
 
     def isdir(self, path):
-        return sync(self.loop, self._isdir, path)
+        return maybe_sync(self._isdir, self, path)
 
     async def _isdir(self, path):
         """Is this entry directory-like?"""
@@ -1063,7 +1067,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
             return False
 
     def exists(self, path):
-        return sync(self.loop, self._exists, path)
+        return maybe_sync(self._exists, self, path)
 
     async def _exists(self, path):
         """Is there a file at the given path"""
@@ -1079,8 +1083,38 @@ class AzureBlobFileSystem(AsyncFileSystem):
             # any exception allowed bar FileNotFoundError?
             return False
 
+    def cat(self, path, recursive=False, on_error="raise", **kwargs):
+        """Fetch (potentially multiple) paths' contents
+        Returns a dict of {path: contents} if there are multiple paths
+        or the path has been otherwise expanded
+        on_error : "raise", "omit", "return"
+            If raise, an underlying exception will be raised (converted to KeyError
+            if the type is in self.missing_exceptions); if omit, keys with exception
+            will simply not be included in the output; if "return", all keys are
+            included in the output, but the value will be bytes or an exception
+            instance.
+        """
+        paths = self.expand_path(path, recursive=recursive)
+        if (
+            len(paths) > 1
+            or isinstance(path, list)
+            or paths[0] != self._strip_protocol(path)
+        ):
+            out = {}
+            for path in paths:
+                try:
+                    out[path] = self.cat_file(path, **kwargs)
+                except Exception as e:
+                    if on_error == "raise":
+                        raise
+                    if on_error == "return":
+                        out[path] = e
+            return out
+        else:
+            return self.cat_file(paths[0])
+
     def expand_path(self, path, recursive=False, maxdepth=None):
-        return sync(self.loop, self._expand_path, path, recursive, maxdepth)
+        return maybe_sync(self._expand_path, self, path, recursive, maxdepth)
 
     async def _expand_path(self, path, recursive=False, maxdepth=None):
         """Turn one or more globs or directories into a list of all matching files"""
@@ -1088,39 +1122,62 @@ class AzureBlobFileSystem(AsyncFileSystem):
             out = await self._expand_path([path], recursive, maxdepth)
         else:
             out = set()
+            path = [self._strip_protocol(p) for p in path]
             for p in path:
                 if has_magic(p):
                     bit = set(await self._glob(p))
                     out |= bit
                     if recursive:
-                        out += await self._expand_path(p)
+                        bit2 = set(await self._expand_path(p))
+                        out |= bit2
                     continue
                 elif recursive:
-                    out |= set(await self._find(p, withdirs=True))
-                # TODO: the following is maybe only necessary if NOT recursive
-                out.add(p)
+                    rec = set(await self._find(p, withdirs=True))
+                    out |= rec
+                if p not in out and (recursive is False or await self._exists(p)):
+                    # only check once for root
+                    out.add(p)
         if not out:
             raise FileNotFoundError
         return list(sorted(out))
 
-    def put(self, lpath, rpath, recursive=False, **kwargs):
-        """Copy file(s) from local.
-        Copies a specific file or tree of files (if recursive=True). If rpath
-        ends with a "/", it will be assumed to be a directory, and target files
-        will go within.
-        Calls put_file for each source.
-        """
-        from fsspec.implementations.local import make_path_posix, LocalFileSystem
+    async def _put_file(self, lpath, rpath, delimiter="/", **kwargws):
+        """ Copy single file to remote """
+        container_name, path = self.split_path(rpath, delimiter=delimiter)
+        cc = self.service_client.get_container_client(container_name)
+        bc = cc.get_blob_client(blob=path)
+        try:
+            with open(lpath, "rb") as f1:
+                await bc.upload_blob(f1)
+            # self.invalidate_cache(self._parent(path))
+            self.invalidate_cache()
+        except ResourceExistsError:
+            raise FileExistsError("File already exists!!")
+        except ResourceNotFoundError:
+            await cc.create_container()
+            await self._put_file(lpath, rpath, delimiter)
+            self.invalidate_cache()
 
-        rpath = self._strip_protocol(rpath)
-        if isinstance(lpath, str):
-            lpath = make_path_posix(lpath)
-        fs = LocalFileSystem()
-        lpaths = fs.expand_path(lpath, recursive=recursive)
-        rpaths = other_paths(lpaths, rpath)
+    put_file = sync_wrapper(_put_file)
 
-        for lpath, rpath in zip(lpaths, rpaths):
-            self.put_file(lpath, rpath, **kwargs)
+    async def _cp_file(self, path1, path2, *kwargs):
+        """ Copy the file at path1 to path2 """
+        # import pdb;pdb.set_trace()
+        container1, path1 = self.split_path(path1, delimiter="/")
+        container2, path2 = self.split_path(path2, delimiter="/")
+
+        cc1 = self.service_client.get_container_client(container1)
+        blobclient1 = cc1.get_blob_client(blob=path1)
+        if container1 == container2:
+            blobclient2 = cc1.get_blob_client(blob=path2)
+        else:
+            cc2 = self.service_client.get_container_client(container2)
+            blobclient2 = cc2.get_blob_client(blob=path2)
+        await blobclient2.start_copy_from_url(blobclient1.url)
+        self.invalidate_cache(container1)
+        self.invalidate_cache(container2)
+
+    cp_file = sync_wrapper(_cp_file)
 
     def upload(self, lpath, rpath, recursive=False, **kwargs):
         """Alias of :ref:`FilesystemSpec.put`."""
@@ -1130,23 +1187,33 @@ class AzureBlobFileSystem(AsyncFileSystem):
         """Alias of :ref:`FilesystemSpec.get`."""
         return self.get(rpath, lpath, recursive=recursive, **kwargs)
 
-    def get(self, rpath, lpath, recursive=False, **kwargs):
-        """Copy file(s) to local.
-        Copies a specific file or tree of files (if recursive=True). If lpath
-        ends with a "/", it will be assumed to be a directory, and target files
-        will go within. Can submit a list of paths, which may be glob-patterns
-        and will be expanded.
-        Calls get_file for each source.
-        """
-        from fsspec.implementations.local import make_path_posix
+    async def _get_file(self, rpath, lpath, recursive=False, delimiter="/", **kwargs):
+        """ Copy single file remote to local """
+        files = await self._ls(rpath)
+        files = [f["name"] for f in files]
+        container_name, path = self.split_path(rpath, delimiter=delimiter)
+        try:
+            cc = self.service_client.get_container_client(container_name)
+            bc = cc.get_blob_client(blob=path)
 
-        rpath = self._strip_protocol(rpath)
-        if isinstance(lpath, str):
-            lpath = make_path_posix(lpath)
-        rpaths = self.expand_path(rpath, recursive=recursive)
-        lpaths = other_paths(rpaths, lpath)
-        for lpath, rpath in zip(lpaths, rpaths):
-            self.get_file(rpath, lpath, **kwargs)
+            if await self._isdir(rpath):
+                os.makedirs(lpath, exist_ok=True)
+            else:
+                with open(lpath, "wb") as my_blob:
+                    stream = await bc.download_blob()
+                    data = await stream.readall()
+                    my_blob.write(data)
+        except Exception as e:
+            raise FileNotFoundError(f"File not found for {e}")
+
+    get_file = sync_wrapper(_get_file)
+
+    def invalidate_cache(self, path=None):
+        if path is None:
+            self.dircache.clear()
+        else:
+            self.dircache.pop(path, None)
+        super(AzureBlobFileSystem, self).invalidate_cache(path)
 
     def _open(
         self,
@@ -1299,27 +1366,33 @@ class AzureBlobFile(io.IOBase):
         ------
         ValueError if none of the connection details are available
         """
-        self.fs.account_url: str = (
-            f"https://{self.fs.account_name}.blob.core.windows.net"
-        )
-        if self.fs.credential is not None:
-            self.container_client = BlobServiceClient(
-                account_url=self.fs.account_url, credential=self.fs.credential
-            ).get_container_client(self.container_name)
-        elif self.fs.connection_string is not None:
-            self.container_client = BlobServiceClient.from_connection_string(
-                conn_str=self.fs.connection_string
-            ).get_container_client(self.container_name)
-        elif self.fs.account_key is not None:
-            self.container_client = BlobServiceClient(
-                account_url=self.fs.account_url, credential=self.fs.account_key
-            ).get_container_client(self.container_name)
-        elif self.fs.sas_token is not None:
-            self.container_client = BlobServiceClient(
-                account_url=self.fs.account_url + self.fs.sas_token, credential=None
-            ).get_container_client(self.container_name)
-        else:
-            raise ValueError("unable to connect with provided params!!")
+        try:
+            self.fs.account_url: str = (
+                f"https://{self.fs.account_name}.blob.core.windows.net"
+            )
+            if self.fs.credential is not None:
+                self.container_client = BlobServiceClient(
+                    account_url=self.fs.account_url, credential=self.fs.credential
+                ).get_container_client(self.container_name)
+            elif self.fs.connection_string is not None:
+                self.container_client = BlobServiceClient.from_connection_string(
+                    conn_str=self.fs.connection_string
+                ).get_container_client(self.container_name)
+            elif self.fs.account_key is not None:
+                self.container_client = BlobServiceClient(
+                    account_url=self.fs.account_url, credential=self.fs.account_key
+                ).get_container_client(self.container_name)
+            elif self.fs.sas_token is not None:
+                self.container_client = BlobServiceClient(
+                    account_url=self.fs.account_url + self.fs.sas_token, credential=None
+                ).get_container_client(self.container_name)
+            else:
+                self.container_client = BlobServiceClient(
+                    account_url=self.fs.account_url
+                ).get_container_client(self.container_name)
+
+        except Exception as e:
+            raise ValueError(f"Unable to connect with provided params for {e}!!")
 
     @property
     def closed(self):
@@ -1397,14 +1470,23 @@ class AzureBlobFile(io.IOBase):
 
     def _initiate_upload(self, **kwargs):
         """Prepare a remote file upload"""
-        self.blob_client = self.container_client.get_blob_client(blob=self.blob)
         self._block_list = []
-        try:
-            self.container_client.delete_blob(self.blob)
-        except ResourceNotFoundError:
-            pass
+        if self.mode == "wb":
+            self.blob_client = self.container_client.get_blob_client(blob=self.blob)
+            try:
+                self.container_client.delete_blob(self.blob)
+            except ResourceNotFoundError:
+                pass
+            else:
+                return self.__initiate_upload()
+        elif self.mode == "ab":
+            self.blob_client = self.container_client.get_blob_client(blob=self.blob)
+            if not self.fs.exists(self.path):
+                self.blob_client.create_append_blob()
         else:
-            return self.__initiate_upload()
+            raise ValueError(
+                "File operation modes other than wb are not yet supported for writing"
+            )
 
     def _upload_chunk(self, final: bool = False, **kwargs):
         """
@@ -1421,12 +1503,28 @@ class AzureBlobFile(io.IOBase):
         length = len(data)
         block_id = len(self._block_list)
         block_id = f"{block_id:07d}"
-        self.blob_client.stage_block(block_id=block_id, data=data, length=length)
-        self._block_list.append(block_id)
+        if self.mode == "wb":
+            try:
+                self.blob_client.stage_block(
+                    block_id=block_id, data=data, length=length
+                )
+                self._block_list.append(block_id)
 
-        if final:
-            block_list = [BlobBlock(_id) for _id in self._block_list]
-            self.blob_client.commit_block_list(block_list=block_list)
+                if final:
+                    block_list = [BlobBlock(_id) for _id in self._block_list]
+                    self.blob_client.commit_block_list(block_list=block_list)
+            except Exception as e:
+                # This step handles the situation where data="" and length=0
+                # which is throws an InvalidHeader error from Azure, so instead
+                # of staging a block, we directly upload the empty blob
+                if block_id == "0000000" and length == 0 and final:
+                    self.blob_client.upload_blob(data=data)
+                else:
+                    raise RuntimeError(f"Failed to upload block with {e}!!")
+        elif self.mode == "ab":
+            self.blob_client.upload_blob(
+                data=data, length=length, blob_type=BlobType.AppendBlob
+            )
 
     def flush(self, force=False):
         """
@@ -1539,9 +1637,17 @@ class AzureBlobFile(io.IOBase):
             return out + [lines[-1]]
         # return list(self)  ???
 
-    def readinto(self, b):
+    def readinto1(self, b):
         return self.readinto(b)
 
+    def readinto(self, b):
+        """mirrors builtin file's readinto method
+        https://docs.python.org/3/library/io.html#io.RawIOBase.readinto
+        """
+        data = self.read(len(b))
+        memoryview(b).cast("B")[: len(data)] = data
+        return len(data)
+
     def read(self, length=-1):
         """
         Return data from cache, or fetch pieces as necessary
@@ -1595,6 +1701,9 @@ class AzureBlobFile(io.IOBase):
         """Whether opened for writing"""
         return self.mode in {"wb", "ab"} and not self.closed
 
+    def __enter__(self):
+        return self
+
     def __exit__(self, *args):
         self.close()
 
